using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using pdftron.Common;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Windows.Controls;

using UIPoint = System.Windows.Point;
using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using pdftron.PDF;


namespace pdftron.PDF.Tools
{
    public class FreehandCreate : SimpleShapeCreate
    {
        protected Path mShape;
        protected List<PDFPoint> mCurrentStrokeList = null; // A list of strokes
        protected List<List<PDFPoint>> mListOfStrokeLists = null; // Contains all the lists above
        protected PathFigureCollection mPathPoints;
        protected PathFigure mCurrentPathFigure;
        protected bool mFirstTime = false; // Whether or not a pointer has been pressed already.
        protected double mLeftmost, mRightmost, mTopmost, mBottommost;
        protected bool mCreateMultiPathAnnotation = false;
        protected int mShapePageNumber = 0;
        protected bool mIgnoreMouseEvents = false;

        public bool CreateMultiPathAnnotation
        {
            get
            {
                return mCreateMultiPathAnnotation;
            }
            set
            {
                mCreateMultiPathAnnotation = value;
            }
        }
        public FreehandCreate(PDFViewWPF view, ToolManager manager)
            : base(view, manager)
        {
            mNextToolMode = ToolManager.ToolType.e_ink_create;
            mToolMode = ToolManager.ToolType.e_ink_create;
            mListOfStrokeLists = new List<List<PDFPoint>>();
        }

        /// <summary>
        /// We need to override the base class and add some functionality here in case we want multiple strokes 
        /// to be part of 1 ink annotation.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        internal override void MouseLeftButtonDownHandler(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            base.MouseLeftButtonDownHandler(sender, e);
            mShapeHasBeenCreated = false;
            if (mFirstTime)
            {
                if (mShapePageNumber != mDownPageNumber)
                {
                    mIgnoreMouseEvents = true;
                }
                else
                {
                    mIgnoreMouseEvents = false;
                }
            }
        }

        /// <summary>
        /// We need to override the base class here, because it will always end the tool.
        /// We want the user to explicitly end simple shape create if they want multiples strokes
        /// to be part of 1 ink annotation.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        internal override void MouseLeftButtonUpHandler(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            // This is what tool does for us.
            mLeftButtonDown = false;
            mIsDragging = false;
            base.MouseLeftButtonUpHandler(sender, e);

            ProcessInputUp(e);
        }

        public override void TouchUpHandler(object sender, System.Windows.Input.TouchEventArgs e)
        {
            base.TouchUpHandler(sender, e);
            ProcessInputUp(e);
        }

        private void ProcessInputUp(System.Windows.Input.InputEventArgs e)
        {
            if (mCreateMultiPathAnnotation)
            {
                mShapeHasBeenCreated = false;
            }
            else if (mIsDrawing)
            {
                Create();
                EndCurrentTool(ToolManager.ToolType.e_pan);
            }
        }

        internal override void KeyDownAction(object sender, System.Windows.Input.KeyEventArgs e)
        {
            if (!mIsDragging)
            {
                if (e.Key == System.Windows.Input.Key.Y)
                {
                    Create();
                    EndCurrentTool(ToolManager.ToolType.e_pan);
                }
                else if (e.Key == System.Windows.Input.Key.N)
                {
                    EndCurrentTool(ToolManager.ToolType.e_pan);
                }
            }
        }


        #region Override Functions

        protected override void Draw()
        {
            if (mIgnoreMouseEvents)
            {
                return;
            }

            double x, y;

            if (!mShapeHasBeenCreated)
            {
                mShapeHasBeenCreated = true;

                x = mDownPoint.X - mPDFView.GetHScrollPos();
                y = mDownPoint.Y - mPDFView.GetVScrollPos();
                mPDFView.ConvScreenPtToPagePt(ref x, ref y, mDownPageNumber);

                if (!mFirstTime)
                {
                    mFirstTime = true;
                    mShapePageNumber = mDownPageNumber; // from now on, mouse presses outside of this page number will be ignored

                    // set bounds
                    mLeftmost = x;
                    mRightmost = x;
                    mTopmost = y;
                    mBottommost = y;
                    mPathPoints = new PathFigureCollection();
                }

                mHalfThickness = mStrokeThickness / 2;
                mCurrentPathFigure = new PathFigure();
                mCurrentPathFigure.StartPoint = new UIPoint(mDownPoint.X - mPageCropOnClient.x1, mDownPoint.Y - mPageCropOnClient.y1);
                mPathPoints.Add(mCurrentPathFigure);
                mListOfStrokeLists.Add(new List<PDFPoint>());
                mCurrentStrokeList = mListOfStrokeLists[mListOfStrokeLists.Count() - 1];

                mCurrentStrokeList.Add(new PDFPoint(x, y));

                PathGeometry geom = new PathGeometry();
                mShape = new Path();
                mShape.StrokeLineJoin = PenLineJoin.Miter;
                mShape.StrokeMiterLimit = 1;
                mShape.Data = geom;
                mShape.StrokeThickness = mDrawThickness;
                mShape.Stroke = mStrokeBrush;
                this.Children.Add(mShape);
            }

            mCurrentPathFigure.Segments.Add(new LineSegment() { Point = new UIPoint(mDragPoint.X - 
                mPageCropOnClient.x1, mDragPoint.Y - mPageCropOnClient.y1) });

            x = mDragPoint.X - mPDFView.GetHScrollPos();
            y = mDragPoint.Y - mPDFView.GetVScrollPos();
            mPDFView.ConvScreenPtToPagePt(ref x, ref y, mDownPageNumber);

            mCurrentStrokeList.Add(new PDFPoint(x, y));

            // Update bounds
            if (x < mLeftmost)
            {
                mLeftmost = x;
            }
            if (x > mRightmost)
            {
                mRightmost = x;
            }
            if (y > mTopmost)
            {
                mTopmost = y;
            }
            if (y < mBottommost)
            {
                mBottommost = y;
            }

            (mShape.Data as PathGeometry).Figures = mPathPoints;
        }

        protected override void Create()
        {
            if (mListOfStrokeLists.Count() == 0)
            {
                return;
            }

            try
            {
                mPDFView.DocLock(true);
                Rect rect = new Rect(mLeftmost, mBottommost, mRightmost, mTopmost);
                if (rect != null)
                {
                    pdftron.PDF.Annots.Ink ink = pdftron.PDF.Annots.Ink.Create(mPDFView.GetDoc().GetSDFDoc(), rect);

                    Annot.BorderStyle bs = ink.GetBorderStyle();
                    bs.width = mStrokeThickness;
                    ink.SetBorderStyle(bs);

                    double sx = mPDFView.GetHScrollPos();
                    double sy = mPDFView.GetVScrollPos();

                    // Shove the points into the Doc
                    int i = 0;
                    PDFPoint pdfp = new PDFPoint();
                    foreach (List<PDFPoint> pointList in mListOfStrokeLists)
                    {
                        int j = 0;
                        foreach (PDFPoint p in pointList)
                        {
                            pdfp.x = p.x;
                            pdfp.y = p.y;
                            ink.SetPoint(i, j, pdfp);
                            j++;
                        }
                        i++;
                    }

                    SetStyle(ink);
                    ink.RefreshAppearance();

                    pdftron.PDF.Page page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    page.AnnotPushBack(ink);

                    mAnnotPageNum = mDownPageNumber;
                    mAnnot = ink;
                    mPDFView.Update(mAnnot, mAnnotPageNum);
                }
            }
            catch (System.Exception ex)
            {
            }
            finally
            {
                mPDFView.DocUnlock();
            }
            mToolManager.RaiseAnnotationAddedEvent(mAnnot);
        }

        protected override void CheckBounds(ref UIPoint point)
        {
            if (mPageCropOnClient != null)
            {
                if (point.X < mPageCropOnClient.x1 + mHalfThickness)
                {
                    point.X = mPageCropOnClient.x1 + mHalfThickness;
                }
                if (point.X > mPageCropOnClient.x2 - mHalfThickness)
                {
                    point.X = mPageCropOnClient.x2 - mHalfThickness;
                }
                if (point.Y < mPageCropOnClient.y1 + mHalfThickness)
                {
                    point.Y = mPageCropOnClient.y1 + mHalfThickness;
                }
                if (point.Y > mPageCropOnClient.y2 - mHalfThickness)
                {
                    point.Y = mPageCropOnClient.y2 - mHalfThickness;
                }
            }
        }

        protected override void HandleResizing()
        {
            if (!mFirstTime)
            {
                return;
            }

            mPageCropOnClient = BuildPageBoundBoxOnClient(mDownPageNumber);
            this.SetValue(System.Windows.Controls.Canvas.LeftProperty, mPageCropOnClient.x1);
            this.SetValue(System.Windows.Controls.Canvas.TopProperty, mPageCropOnClient.y1);
            this.Width = mPageCropOnClient.x2 - mPageCropOnClient.x1;
            this.Height = mPageCropOnClient.y2 - mPageCropOnClient.y1;

            double sx = mPDFView.GetHScrollPos();
            double sy = mPDFView.GetVScrollPos();
            double x, y;

            mLeftmost = -1;
            mRightmost = -1;
            mTopmost = -1;
            mBottommost = -1;

            mZoomLevel = mPDFView.GetZoom();
            mDrawThickness = mStrokeThickness * mZoomLevel;
            mHalfThickness = mDrawThickness / 2;

            this.Children.Clear();

            PathGeometry geom = new PathGeometry();
            mShape = new Path();
            mShape.StrokeLineJoin = PenLineJoin.Miter;
            mShape.StrokeMiterLimit = 1;
            mShape.Data = geom;
            mShape.StrokeThickness = mDrawThickness;
            mShape.Stroke = mStrokeBrush;
            this.Children.Add(mShape);

            mPathPoints = new PathFigureCollection();

            foreach (List<PDFPoint> strokeList in mListOfStrokeLists)
            {
                mCurrentPathFigure = new PathFigure();
                mPathPoints.Add(mCurrentPathFigure);
                bool first = true;
                foreach (PDFPoint point in strokeList)
                {
                    x = point.x;
                    y = point.y;
                    mPDFView.ConvPagePtToScreenPt(ref x, ref y, mDownPageNumber);

                    if (first)
                    {
                        first = false;
                        mCurrentPathFigure.StartPoint = new UIPoint(x + sx - mPageCropOnClient.x1, y + sy - mPageCropOnClient.y1);
                    }
                    else
                    {
                        mCurrentPathFigure.Segments.Add(new LineSegment() { Point = new UIPoint(x + sx - 
                            mPageCropOnClient.x1, y + sy - mPageCropOnClient.y1) });
                    }

                }
            }
            (mShape.Data as PathGeometry).Figures = mPathPoints;
        }

        protected override void MultiTouchDetected()
        {
            if (mPathPoints.Contains(mCurrentPathFigure))
            {
                mPathPoints.Remove(mCurrentPathFigure);
            }
            if (mListOfStrokeLists.Contains(mCurrentStrokeList))
            {
                mListOfStrokeLists.Remove(mCurrentStrokeList);
            }
        }

        #endregion Override Functions


        #region Context Menu

        public override void AddContextMenuItems(System.Windows.Controls.ContextMenu menu, double x, double y)
        {
            base.AddContextMenuItems(menu, x, y);

            // We're drawing a shape
            if (mFirstTime)
            {
                Separator sep = new Separator();
                menu.Items.Add(sep);

                MenuItem m = new MenuItem();
                m.Header = "Save Shape";
                m.Click += ContextMenu_Finish;
                menu.Items.Add(m);


                m = new MenuItem();
                m.Header = "Discard Shape";
                m.Click += ContextMenu_Discard;
                menu.Items.Add(m);
            }
        }

        private void ContextMenu_Finish(object sender, System.Windows.RoutedEventArgs e)
        {
            SaveShape();
        }

        private void ContextMenu_Discard(object sender, System.Windows.RoutedEventArgs e)
        {
            DiscardShape();
        }


        #endregion Context Menu


        /// <summary>
        /// Save the current drawing
        /// </summary>
        public void SaveShape()
        {
            Create();
            EndCurrentTool(ToolManager.ToolType.e_pan);
        }

        // Discard the current drawing
        public void DiscardShape()
        {
            EndCurrentTool(ToolManager.ToolType.e_pan);
        }
    }
}
